import peewee


__all__ = ('Handle',)


class Handle:

    __slots__ = ('_seed', '_authorize', '_error')

    def __init__(self, seed, authorize, error):

        self._seed = seed

        self._authorize = authorize

        self._error = error

    _low = 0

    async def _execute(self, accept, action, request):

        self._authorize(request)

        try:

            payload = await request.json()

        except:

            raise self._error('invalid data')

        try:

            (keys, *others) = payload

        except (TypeError, ValueError):

            raise self._error('invalid payload')

        if not isinstance(keys, list):

            raise self._error('invalid keys')

        if accept and len(others) > self._low:

            data = others[self._low]

            if not isinstance(data, dict):

                raise self._error('invalid data', index)

            act = action(keys, **data)

        else:

            act = action(keys)

        try:

            data = await act

        except peewee.DatabaseError as error:

            raise self._error('database error', str(error))

        return data

    _get = False

    def get(self, request):

        return self._execute(self._get, self._seed.get, request)

    _create = True

    def create(self, request):

        return self._execute(self._create, self._seed.create, request)

    _update = True

    def update(self, request):

        return self._execute(self._update, self._seed.update, request)

    _delete = False

    def delete(self, request):


        return self._execute(self._delete, self._seed.delete, request)



import asyncio
import aiohttp.web
import functools

from . import access
from . import seed
from . import handle


__all__ = ()


def authorization(error, key, value, store):

    try:

        provide = store[key]

    except KeyError:

        raise error('missing')

    if provide == value:

        return

    raise error('invalid')


def fail(cls, *data):

    response = aiohttp.web.json_response(data)

    return cls(body = response.body, content_type = response.content_type)


authorization_header = 'Authorization'


authorization_error = aiohttp.web.HTTPUnauthorized


authorization_fail = functools.partial(fail, authorization_error)


authorization_parts = (authorization_fail, authorization_header)


specific_authorize = functools.partial(authorization, *authorization_parts)


def specific_authorization(spot, value, request):

    store = getattr(request, spot)

    specific_authorize(value, store)


authorization_spot = 'headers'


authorize = functools.partial(specific_authorization, authorization_spot)


authorizer = lambda value: functools.partial(authorize, value)


handle_error = aiohttp.web.HTTPBadRequest


handle_fail = functools.partial(fail, authorization_error)


ignore = ('GET',)


methods = (*ignore, 'POST', 'PATCH', 'DELETE')


def listen(alert, handle):

    async def wrapper(request):

        data = await handle(request)

        await alert(data)

        return aiohttp.web.json_response(data)

    return wrapper


silent = asyncio.coroutine(lambda *args: None)


def setup(client, objects, model, authorize, fail, parse = tuple):

    _access = access.Access(objects, model)

    client.add_info(_access.name, _access.primary)

    _seed = seed.Seed(_access, parse)

    _handle = handle.Handle(_seed, authorize, fail)

    sub_handlers = (_handle.get, _handle.create, _handle.update, _handle.delete)

    controller = '/' + _access.name

    def vocal(*args):

        async def wrapper(data):

            await client.dispatch(*args, data)

        return wrapper

    for (method, sub_handler) in zip(methods, sub_handlers):

        dispatch = silent if method in ignore else vocal(method, _access.name)

        handler = listen(dispatch, sub_handler)

        yield (method, controller, handler)





__all__ = ('Seed',)


class Seed:

    __slots__ = ('_access', '_parse')

    def __init__(self, access, parse):

        self._access = access

        self._parse = parse

    async def get(self, keys):

        data = await self._access.get(keys)

        return self._parse(data)

    async def create(self, keys, **data):

        await self._access.create(keys, **data)

        data = await self._access.get(keys)

        return self._parse(data)

    async def update(self, keys, **data):

        await self._access.update(keys, **data)

        data = await self._access.get(keys)

        return self._parse(data)

    async def delete(self, keys):

        data = await self._access.get(keys)

        await self._access.delete(keys)

        return self._parse(data)



import operator
import functools


__all__ = ('Access',)


class Access:

    __slots__ = ('_objects', '_model')

    def __init__(self, objects, model):

        self._objects = objects

        self._model = model

    @property
    def name(self):

        return self._model._meta.name

    @property
    def primary(self):

        return self._model._meta.primary_key.field_names

    def query(self, keys):

        fields = (getattr(self._model, key) for key in self.primary)

        associations = (field == key for (field, key) in zip(fields, keys))

        action = functools.reduce(operator.and_, associations)

        return action

    def _execute(self, action, keys):

        if keys:

            query = self.query(keys)

            action = action.where(query)

        return self._objects.execute(action)

    def _revert(self, data):

        for (key, value) in data.items():

            if not value is None:

                continue

            data[key] = getattr(self._model, key).default

    def get(self, keys):

        action = self._model.select().dicts()

        return self._execute(action, keys)

    def create(self, keys, **data):

        self._revert(data)

        extra = zip(self.primary, keys)

        data.update(extra)

        action = self._model.insert(**data)

        return self._execute(action, None)

    def update(self, keys, **data):

        self._revert(data)

        action = self._model.update(**data)

        return self._execute(action, keys)

    def delete(self, keys):

        action = self._model.delete()

        return self._execute(action, keys)
